/^Project-Id-Version:/ b
/^Report-Msgid-Bugs-To:/ b
/^POT-Creation-Date:/ b
/^PO-Revision-Date:/ b
/^Last-Translator:/ b
/^Language-Team:/ b
/^Language:/ b
/^MIME-Version:/ b
/^Content-Type:/ b
/^Content-Transfer-Encoding:/ b
/^Plural-Forms:/ b

s,%\([0-9]*\)C,%\1#,g

s,a,α,g
s,b,ϭ,g
s,c,ϲ,g
s,d,δ,g
s,e,ε,g
s,f,φ,g
s,g,γ,g
s,h,χ,g
s,i,ι,g
s,j,ϊ,g
s,k,κ,g
s,l,λ,g
s,m,μ,g
s,n,ν,g
s,o,ο,g
s,p,π,g
s,q,ϗ,g
s,r,ρ,g
s,s,σ,g
s,t,τ,g
s,u,υ,g
s,v,β,g
s,w,ω,g
s,x,ξ,g
s,y,ϋ,g
s,z,ζ,g
s,A,Α,g
s,B,Ϭ,g
s,C,ϲ,g
s,D,Δ,g
s,E,Ε,g
s,F,Φ,g
s,G,Γ,g
s,H,Χ,g
s,I,Ι,g
s,J,Ϊ,g
s,K,Κ,g
s,L,Λ,g
s,M,Μ,g
s,N,Ν,g
s,O,Ο,g
s,P,Π,g
s,Q,Ϗ,g
s,R,Ρ,g
s,S,Σ,g
s,T,Τ,g
s,U,Υ,g
s,V,Β,g
s,W,Ω,g
s,X,Ξ,g
s,Y,Ϋ,g
s,Z,Ζ,g
s,А,А,g
s,Б,Б,g
s,Ц,Ц,g
s,Д,Д,g
s,Е,Е,g
s,Ф,Ф,g
s,Г,Г,g
s,Х,Х,g
s,И,И,g
s,Й,Й,g
s,К,К,g
s,Л,Л,g
s,М,М,g
s,Н,Н,g
s,О,О,g
s,П,П,g
s,Ќ,Ќ,g
s,Р,Р,g
s,С,С,g
s,Т,Т,g
s,У,У,g
s,В,В,g
s,Ў,Ў,g
s,Ѯ,Ѯ,g
s,Ѝ,Ѝ,g
s,З,З,g

s,%\([0-9]*\)#,%\1C,g
s,%\([0-9]*\)ϲ,%\1c,g
s,%\([0-9]*\)δ,%\1d,g
s,%\([0-9]*\)λδ,%\1ld,g
s,%\([0-9]*\)λλδ,%\1lld,g

s,%\([0-9\.\*]*\)σ,%\1s,g
s,%\([0-9\.\*]*\)μ,%\1m,g
s,%\([0-9]*\)λυ,%\1lu,g
s,%\([0-9]*\)υ,%\1u,g
s,%\([0-9]*\)λλυ,%\1llu,g
s,%\([0-9]*\)ξ,%\1x,g
s,%\([0-9]*\)λξ,%\1lx,g
s,%\([0-9]*\)λλξ,%\1llx,g
